ChainTable = {};
	
---------------------------------------------------------------
-- Checks if all prerequisite specializations are loaded
---------------------------------------------------------------
function ChainTable.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations)
end

---------------------------------------------------------------
--
---------------------------------------------------------------
function ChainTable.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", ChainTable)--[[
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", ChainTable)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", ChainTable)]]
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", ChainTable)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", ChainTable)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", ChainTable)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", ChainTable)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", ChainTable)
end

---------------------------------------------------------------
--
---------------------------------------------------------------
function ChainTable.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier",	ChainTable.getDirtMultiplier)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier",	ChainTable.getWearMultiplier)
    --SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeTurnedOn",		ChainTable.getCanBeTurnedOn)
end
	
---------------------------------------------------------------
-- Called on loading
---------------------------------------------------------------
function ChainTable:onLoad(savegame)
    local spec = self.spec_chainTable
    ChainTable.loadChainTable(self, spec, self.xmlFile, self.components)
    local moveColDisableCollisionPairs = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.chainTable#moveColDisableCollisionPairs"), true)
    if moveColDisableCollisionPairs then
        for _, component in pairs(self.components) do
            for _, node in pairs(spec.moveColNodes) do
                setPairCollision(component.node, node, false)
            end
        end
    end
	self.turnOffTimer = 0
	--print (self.turnOffTimer)
end

--[[-------------------------------------------------------------
-- Called on Update
---------------------------------------------------------------
function ChainTable:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        local spec = self.spec_chainTable
        if streamReadBool(streamId) then
            spec.crushingTime = 1000
        else
            spec.crushingTime = 0
        end
    end
end

---------------------------------------------------------------
-- Called on Update
---------------------------------------------------------------
function ChainTable:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_chainTable
        streamWriteBool(streamId, spec.crushingTime > 0)
    end
end

---------------------------------------------------------------]]
-- Called on deleting
---------------------------------------------------------------
function ChainTable:onDelete()
    ChainTable.deleteChainTable(self, self.spec_chainTable)
end

---------------------------------------------------------------
-- Called on Update
---------------------------------------------------------------
function ChainTable:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    ChainTable.updateChainTable(self, self.spec_chainTable, dt, self:getIsTurnedOn())
end

---------------------------------------------------------------
-- Called on Update Tick
---------------------------------------------------------------
function ChainTable:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_chainTable
    -- turn on/off automatically if tree node is in trigger
	if self.isServer then
		if spec.turnOnAutomatically then
			if self.turnOffTimer > 0 then
				self.turnOffTimer = self.turnOffTimer - dt
				--print(self.turnOffTimer)
			end
			if next(spec.moveTriggerNodes) ~= nil then
				if not self:getIsTurnedOn() then
					self:setIsTurnedOn(true)
					self.turnOffTimer = 5000
					--print ("ON")
				end
			elseif self.turnOffTimer < 0 then
				--print ("OFF")
				self:setIsTurnedOn(false)
				self.turnOffTimer = 0
			end
		end
	end
end

---------------------------------------------------------------
-- Called on turn on
---------------------------------------------------------------
function ChainTable:onTurnedOn()
    ChainTable.turnOnChainTable(self, self.spec_chainTable)
    --Lights.setBeaconLightsVisibility(self, true)
    Lights.setTurnLightState(self, Lights.TURNLIGHT_HAZARD)
end

---------------------------------------------------------------
-- Called on turn off
---------------------------------------------------------------
function ChainTable:onTurnedOff()
    ChainTable.turnOffChainTable(self, self.spec_chainTable)
    --Lights.setBeaconLightsVisibility(self, false)
    Lights.setTurnLightState(self, Lights.TURNLIGHT_OFF)
end


--[[-------------------------------------------------------------
-- Action LoadOut
---------------------------------------------------------------
function ChainTable:onRegisterActionEvents()

	 print(string.format("TEST Register action events"))

	if self.isClient then

		if self.ActionEventIds ~= nil then
			for _, eventId in pairs(self.ActionEventIds) do
				g_inputBinding:removeActionEvent(eventId)
			end
		end

		self.ActionEventIds = {}

		for actionName, actionDetails in pairs(ChainTable.Actions) do
			local internalActionName = actionDetails[1]
			local contextName = actionDetails[2]
			local priority = actionDetails[3]

			if self.isClient then
				 print(string.format("TEST Registering action: %s ...", actionName))
				g_inputBinding:beginActionEventsModification(contextName)

				local colliding = false
				_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
				if colliding then
				 print(string.format("Error: ChainTable got a colliding action: %s", actionName))
				end

				local actionEventId = ""
				_, actionEventId = g_inputBinding:registerActionEvent(actionName, self, ChainTable.actionEventCallback, false)
				g_inputBinding:setActionEventText(actionEventId, g_i18n:getText(string.format("action_%s", actionName)))
				g_inputBinding:setActionEventTextPriority(actionEventId, priority)
				g_inputBinding:setActionEventTextVisibility(actionEventId, true)

				self.ActionEventIds[actionName] = actionEventId

				g_inputBinding:endActionEventsModification()
			end
		end

	end
end

function ChainTable.actionEventCallback(self, actionName, inputValue, callbackState)

	 print("TEST ChainTable action callback: ", actionName, inputValue, callbackState)

	local isTriggered = (inputValue > 0)
	self.TriggeredActions[ChainTable.Actions[actionName][1]]-- = isTriggered
--end

--[[-------------------------------------------------------------
-- Returns if user is allowed to turn on the vehicle
---------------------------------------------------------------
function ChainTable:getCanBeTurnedOn(superFunc)
    local spec = self.spec_chainTable
    if spec.turnOnAutomatically then
		return false
    end
    return superFunc(self)
end

---------------------------------------------------------------]]
-- Returns current dirt multiplier
---------------------------------------------------------------
function ChainTable:getDirtMultiplier(superFunc)
    local multiplier = superFunc(self)
	if self.turnOffTimer ~= nil then
		if self.turnOffTimer > 0 then
			multiplier = multiplier + self:getWorkDirtMultiplier()
		end
	end
    return multiplier
end

---------------------------------------------------------------
-- Returns current wear multiplier
---------------------------------------------------------------
function ChainTable:getWearMultiplier(superFunc)
    local multiplier = superFunc(self)
    if self.turnOffTimer ~= nil then
		if self.turnOffTimer > 0 then
			multiplier = multiplier + self:getWorkWearMultiplier()
		end
	end
    return multiplier
end

---------------------------------------------------------------
-- Load Table from xml file
---------------------------------------------------------------
function ChainTable.loadChainTable(self, chainTable, xmlFile, rootNode)
    chainTable.vehicle = self
    chainTable.chainTableMoveTriggerCallback = ChainTable.chainTableMoveTriggerCallback
    chainTable.playerTriggerCallback = ChainTable.playerTriggerCallback
    local xmlRoot = getXMLRootName(xmlFile)
    XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, xmlRoot .. ".chainTable.moveTrigger(0)#index", xmlRoot .. ".chainTable.moveTriggers.trigger#node") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, xmlRoot .. ".chainTable.playerTrigger(0)#index", xmlRoot .. ".chainTable.playerTriggers.trigger#node")
	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, xmlRoot .. ".chainTable.moveCollision(0)#index", xmlRoot .. ".chainTable.moveCollisions.collision#node") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, xmlRoot .. ".turnedOnRotationNodes.turnedOnRotationNode#type", xmlRoot .. ".chainTable.animationNodes.animationNode", "chainTable") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, xmlRoot .. ".turnedOnScrollers.turnedOnScroller", xmlRoot .. ".chainTable.animationNodes.animationNode") --FS17 to FS19
    local baseKey = xmlRoot .. ".chainTable"
    chainTable.moveTriggers = {}
    local i = 0
    while true do
        local key = string.format("%s.moveTriggers.trigger(%d)", baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end
        local node = I3DUtil.indexToObject(rootNode, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)
        if node ~= nil then
            table.insert(chainTable.moveTriggers, node)
        end
        i = i + 1
    end
	chainTable.moveColNodes = {}
    i = 0
    while true do
        local key = string.format("%s.moveCollisions.collision(%d)", baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end
        local node = I3DUtil.indexToObject(rootNode, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)
        if node ~= nil then
            table.insert(chainTable.moveColNodes, node)
        end
        i = i + 1
    end
    chainTable.moveVelocityZ = Utils.getNoNil(getXMLFloat(xmlFile, baseKey.."#moveVelocityZ"), 0.5) -- m/s
    chainTable.moveMaxForce = Utils.getNoNil(getXMLFloat(xmlFile, baseKey.."#moveMaxForce"), 7) -- input is kN
    chainTable.moveTriggerNodes = {}
    if self.isServer and chainTable.moveTriggers ~= nil then
		for _,node in pairs(chainTable.moveTriggers) do
            addTrigger(node, "chainTableMoveTriggerCallback", chainTable)
        end
    end
    chainTable.turnOnAutomatically = Utils.getNoNil(getXMLBool(xmlFile, baseKey .. "#automaticallyTurnOn"), true)
    if self.isClient then
        chainTable.animationNodes = g_animationManager:loadAnimations(xmlFile, baseKey..".animationNodes", rootNode, self, self.i3dMappings)
    end
	chainTable.playerTriggers = {}
	local i = 0
    while true do
        local key = string.format("%s.playerTriggers.trigger(%d)", baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end
        local node = I3DUtil.indexToObject(rootNode, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)
        if node ~= nil then
            table.insert(chainTable.playerTriggers, node)
		else
			g_logManager:error("[ChainTable] - PlayerTriggerNodes is NIL ",  self.configFileName);
        end
        i = i + 1
    end
    chainTable.playerTriggerNodes = {}
    if self.isServer and chainTable.playerTriggers ~= nil then
		for _,node in pairs(chainTable.playerTriggers) do
			addTrigger(node, "playerTriggerCallback", chainTable)
		end
	end
end

---------------------------------------------------------------
-- Delete Table
---------------------------------------------------------------
function ChainTable.deleteChainTable(self, chainTable)
    if self.isServer and chainTable.moveTriggers ~= nil then
        for _,node in pairs(chainTable.moveTriggers) do
            removeTrigger(node)
        end
    end
	if self.isServer and chainTable.playerTriggers ~= nil then
        for _,node in pairs(chainTable.playerTriggers) do
            removeTrigger(node)
        end
    end
    if self.isClient then
        g_animationManager:deleteAnimations(chainTable.animationNodes)
    end
end

---------------------------------------------------------------
-- Update Table
---------------------------------------------------------------
function ChainTable.updateChainTable(self, chainTable, dt, isTurnedOn)
    if isTurnedOn then
        if self.isServer then
            for id in pairs(chainTable.moveTriggerNodes) do
                if not entityExists(id) then
                    chainTable.moveTriggerNodes[id] = nil
				end
            end
            if next(chainTable.moveTriggerNodes) ~= nil then
                self:raiseActive()
            end
        end
    end
end

---------------------------------------------------------------
-- Turn on Table
---------------------------------------------------------------
function ChainTable.turnOnChainTable(self, chainTable)
    if self.isServer and chainTable.moveColNodes ~= nil then
        for _,node in pairs(chainTable.moveColNodes) do
            setFrictionVelocity(node, chainTable.moveVelocityZ)
        end
    end
    if self.isClient then
        chainTable.isWorkSamplePlaying = false
		g_animationManager:startAnimations(chainTable.animationNodes)
    end
end

---------------------------------------------------------------
-- Turn off Table
---------------------------------------------------------------
function ChainTable.turnOffChainTable(self, chainTable)
    if self.isServer then
		if chainTable.moveColNodes ~= nil then
			for _,node in pairs(chainTable.moveColNodes) do
				setFrictionVelocity(node, 0.0)
			end
		end
	end
    if self.isClient then
        g_effectManager:stopEffects(chainTable.crushEffects)
        chainTable.isWorkSamplePlaying = false
        if self.isClient then
			g_animationManager:stopAnimations(chainTable.animationNodes)
		end
    end
end


---------------------------------------------------------------
-- MoveTrigger callback
---------------------------------------------------------------
function ChainTable.chainTableMoveTriggerCallback(self, triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    local splitType = g_splitTypeManager:getSplitTypeByIndex(getSplitType(otherActorId))
	if splitType ~= nil then
		if onEnter then
			self.moveTriggerNodes[otherActorId] = Utils.getNoNil(self.moveTriggerNodes[otherActorId],0)+1
            self.vehicle:raiseActive()
		elseif onLeave then
            local c = self.moveTriggerNodes[otherActorId]
            if c ~= 0 then
                c = c-1
                if c == 0 then
                    self.moveTriggerNodes[otherActorId] = nil
                else
                    self.moveTriggerNodes[otherActorId] = c
                end
            end
		end
	end
end

---------------------------------------------------------------
-- PlayerTrigger Call back
---------------------------------------------------------------

function ChainTable.playerTriggerCallback(self, triggerId, otherId, onEnter, onLeave, onStay)
	if (g_currentMission.controlPlayer and g_currentMission.player and otherId == g_currentMission.player.rootNode) then
		if (onEnter) then
			self.TriggerActive = true
			self:raiseActive()
		elseif (onLeave) then
			self.TriggerActive = false
			self.vehicle:raiseActive()
        end
	end
end
